import json

class default_server:
    def __init__(self):
        pass

    def send_header(self, sock, header):
        for field in header:
            sock.write(bytes(field + ": " + header[field] + "\r\n", "utf-8"))

    def send_response(self, sock, code=200, header={}, data=None):
        if code == 200: sock.write(b"HTTP/1.1 200 OK\r\n")
        if code == 201: sock.write(b"HTTP/1.1 201 CREATED\r\n")
        if code == 202: sock.write(b"HTTP/1.1 202 ACCEPTED\r\n")
        if code == 204: sock.write(b"HTTP/1.1 204 NO CONTENT\r\n")
        if code == 400: sock.write(b"HTTP/1.1 400 BAD REQUEST\r\n")
        if code == 404: sock.write(b"HTTP/1.1 404 NOT FOUND\r\n")
        if code == 405: sock.write(b"HTTP/1.1 405 METHOD NOT ALLOWED\r\n")

        b_data = b""

        if type(data) is str:
            header["Content-Type"] = "text/plain; charset=UTF-8"
            b_data = bytes(data, "utf-8")
        else:
            if data is not None:
                header["Content-Type"] = "application/json; charset=UTF-8"
                b_data = bytes(json.dumps(data), "utf-8")

        if data is not None:
            header["Content-Length"] = str(len(b_data))

            header["Access-Control-Allow-Origin"] = "http://localhost"
            header["Access-Control-Allow-Headers"] =  "Origin, X-Requested-With, Content-Type, Accept"
            header["Connection"] = "close"

        self.send_header(sock, header)
        
        if data is not None:
            sock.write(b"\r\n")
            sock.write(b_data)


import ep_default_server
import re
import json
import ep_config
import ep_logging

class rest_server(ep_default_server.default_server):
    def __init__(self, logger=None):
        super().__init__()
        self.logger = logger if logger is not None else ep_logging.default_logger(appname="http_rs")

    def serve(self, sock, request):
        m = re.match(request["route"], request["ressource"])
        path = m.group(1)
        if path.endswith("/"):
            path = path[0:-1]

        if request["method"] == "GET":
            self.get(path, request, sock)
        
        if request["method"] == "PUT":
            self.put(path, request, sock)
        
        if request["method"] == "OPTIONS":
            self.options(path, request, sock)


class config_rest_server(rest_server):
    def __init__(self, config_file="config.json", logger=None):
        super().__init__(logger=logger)
        self.config_file = config_file            
    
    def get(self, path, req, sock):
        config = ep_config.config(self.config_file)
        config.load()
        data = config.get(path)
        self.send_response(sock, code=200, data=data)

    def put(self, path, req, sock):
        config = ep_config.config(self.config_file)
        config.load()
        created = not config.has(path)
        config.set(path, req["fields"])
        config.save()
        if created:
            self.send_response(sock, 201)
        else:
            self.send_response(sock, 200)

    def options(self, path, req, sock):
        header = {"Access-Control-Allow-Methods": "PUT, POST, GET, DELETE, OPTIONS"}
        self.send_response(sock, 200, header=header)


class sensor_rest_server(rest_server):
    def __init__(self, routes, logger=None):
        super().__init__(logger=logger)
        self.routes = routes

    def get(self, path, req, sock):
        reading = ""
        code = 404

        for route in self.routes:
            self.logger.debug("Used route: " + route[0])
            g = re.match(route[0], path)
            if g is not None:
                reading = route[1](path)
                code = 200
                break
                
        self.send_response(sock, code=code, data=reading)
        


import ep_default_server
import os
import re
import ep_logging

class file_server(ep_default_server.default_server):
    def __init__(self, html_dir, default_file="config.html", logger=None):
        super().__init__()
        self.html_dir = html_dir
        self.default_file = default_file
        self.logger = logger if logger is not None else ep_logging.default_logger(appname="http_fs")

    def serve(self, sock, request):
        if "?" in request["ressource"]:
            file, _ = request["ressource"].split("?")
        else:
            file = request["ressource"]

        m = re.match(request["route"], file)
        sock.write(b"HTTP/1.1 200 OK\r\n")  

        header = {
            "Content-Type": "text/html; charset=UTF-8",
        }
        header["Connection"] = "close"
            
        self.send_header(sock, header)
        sock.write(b"\r\n")
        self.return_file(m.group(1), sock)

    def return_file(self, file, sock):
        self.logger.debug("Requested file: " + file)
        if file in ["", "/"]:
            file = self.default_file
        if file == "favicon.ico":
            file = "favicon.png"
        self.logger.info("Served file: " + file)
        if file in os.listdir(self.html_dir):
            with open(self.html_dir + file, "rb") as f:
                data = f.read(512)
                while data != b"":
                    sock.send(data)
                    data = f.read(512)


import socket
import re
import os
import json
import _thread
import time
import select
import ep_logging

class http_server:
    def __init__(self, micropython_optimize=False, routes=[], logger=None):
        self.micropython_optimize = micropython_optimize
        self.routes = routes
        self.logger = logger if logger is not None else ep_logging.default_logger(appname="http")

    def start(self):
        _thread.start_new_thread(self._start, ())

    def _start(self):
        self.logger.info("Server started")
        s = socket.socket()
        ai = socket.getaddrinfo("0.0.0.0", 80)
        addr = ai[0][-1]
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        s.bind(addr)
        s.listen(5)
        self.logger.debug("listening...")
        while True:
            self.logger.debug("Waiting for new Connection")
            client_sock, client_addr = s.accept()
            self.logger.info("New client: " + str(client_addr))
            
            if not self.micropython_optimize:
                client_stream = client_sock.makefile("rwb")
            else:
                client_stream = client_sock
            
            _thread.start_new_thread(self.handle_client, (client_stream, client_sock))
    
    def handle_client(self, client_stream, client_sock):
        try:
            self.logger.debug("Reading Request")
            req = client_stream.readline()
            
            self.logger.debug("Processing Request")
            req = self.split_request(req)
            if req is not None:
                self.process_req(req, client_stream)
        except Exception as e:
            print('--- Caught Exception ---')
            import sys
            sys.print_exception(e)
            self.logger.error("error")
            print('----------------------------')
        finally:
            client_stream.close()
            if not self.micropython_optimize:
                client_sock.close()
            self.logger.debug("Socked closed")

    def process_req(self, req, sock):
        req["header"] = header = self.split_header(sock)
        
        req["query"]  = ""
        if "?" in req["ressource"]:
            _, req["query"]  = req["ressource"].split("?")
        
        if req["method"] in ["POST", "PATCH", "PUT"]:
            req["query"] = sock.read(int(header["Content-Length"])).decode("utf8")
    
        req["fields"] = self.split_fields(req["query"], encoding=header.get("Content-Type", "application/x-www-form-urlencoded"))
        
        for route in self.routes:
            if re.match(route[0], req["ressource"]):
                req["route"] = route[0]
                self.logger.debug(req)
                route[1](sock, req)
                return
        else:
            self.logger.warning("no matching route found")

    def split_header(self, sock):
        header = {}
        while True:
            h = sock.readline().decode("utf8")
            if h == "" or h == "\r\n":
                break
            if ":" in h:
                key, val = h.split(": ")
                header[key] = val[:-2]
        return header


    def split_fields(self, query, encoding="application/x-www-form-urlencoded"):
        self.logger.debug("splitting fields, encoding is " + encoding)
        fields = {}
        if encoding == "application/x-www-form-urlencoded":
            if "=" in query:
                for field in query.split("&"):
                    key, val = field.split("=")
                    fields[key] = self.urlencode(val)
        elif encoding.startswith("text/plain"):
            fields = query
        elif encoding.startswith("application/json"):
            fields = json.loads(query)
        return fields


    def split_request(self, req):
        self.logger.debug("Splitting Request: " + str(req))
        g = re.match(r'(GET|POST|PUT|OPTIONS) (.*?) HTTP\/(.*)', req.decode("utf8"))
        if g is not None:
            request = {
                "method": g.group(1),
                "ressource": g.group(2)[1:],
                "version": g.group(3)[:-2]
            }
        else:
            request = None
            self.logger.warning("Invalid Request")
        
        return request               

    def urlencode(self, line):
        d = {
            r"+": " ",
            r"%20": "!", r"%21": "!", r"%22": '"', r"%23": "#", r"%24": "$", r"%25": "%", r"%26": "&", r"%27": "'",
            r"%28": "(", r"%29": ")", r"%2A": "*", r"%2B": "+", r"%2C": ",", r"%2D": "-", r"%2E": ".", r"%2F": "/",
            r"%3A": ":", r"%3B": ";", r"%3C": "<", r"%3D": "=", r"%3E": ">", r"%3F": "?",
            r"%4A": "@", r"%5B": "[", r"%5C": "\\", r"%5D": "]", r"%7B": "{", r"%7C": "|", r"%7D": "}",
        }
        for key in d:
            line = line.replace(key, d[key])
        return line



